(* non-embedded, overloading, folding *)

use "../lib/smlprelude";

(* for overloading *)
fun new_method f =
	let val M = ref f
	    fun get () = !M
	    fun over f = M := f (get ())
	in (over,get)
	end;



(* root objects -------------------------------------------------------	*)

open ExtensibleClosed;   (* extensible closed datatype, % : closed *)

signature ROOT =
sig
    exception Wrap
    type 'a subself
    type self = closed subself
    type wrap
    type 'a packet = 'a * ('a -> wrap) * (wrap -> 'a);
    val wrap'self : self -> wrap
    val self'wrap : wrap -> self
    val unpack : 'a subself -> ('b -> 'b subself) * 'a
    val pack : unit -> 'a -> 'a subself
    val new : unit -> self packet
end;

structure Root :> ROOT =
struct

(* types *)
type 'a subself = 'a;
type self = closed subself;
type wrap = self;
type 'a packet = 'a * ('a -> wrap) * (wrap -> 'a);

(* packing *)
fun unpack ext = (I,ext);
fun pack () ext = ext;

(* wrapping *)
exception Wrap;
fun wrap'self p = p;
fun self'wrap p = p;

(* new *)
fun new () = (pack () %, wrap'self, self'wrap);
end;
 


(* eqobj -------------------------------------------------------------- *)

signature EQOBJ =
sig
    type 'a ext
    type 'a subself = 'a ext Root.subself
    type self = closed subself
    val wrap'self : self -> Root.wrap
    val self'wrap : Root.wrap -> self
    val unpack : 'a subself -> ('b -> 'b subself) * 'a
    val pack : unit -> 'a -> 'a subself
    exception Eq
    val eq : 'a subself Root.packet -> 'b subself Root.packet -> bool
    val overload_eq : ((Root.wrap -> Root.wrap -> bool) ->
		       (Root.wrap -> Root.wrap -> bool)) -> unit
    val new : unit -> self Root.packet
end;

structure Eqobj :> EQOBJ =
struct

(* types *)
type 'a ext = unit * 'a;
type 'a subself = 'a ext Root.subself;
type self = closed subself;

(* packing *)
fun unpack p = explode (Root.unpack p);
fun repack f p = implode f (Root.unpack p);
fun pack () ext = Root.pack () ((), ext);

(* wrapping *)
val (Self,matches) = create (None : closed ext option);
fun wrap'self p =
	let val (f,ext) = Root.unpack p in Root.wrap'self (f (Self ext)) end;
fun self'wrap w =
	let val (f,ww) = Root.unpack (Root.self'wrap w)
	in case matches ww of Some ext => f ext | _ => raise Root.Wrap
	end;

(* messages *)
exception Eq;
val (overload_eq, get_eq) =
	let fun eqeq (p : Root.wrap) (q : Root.wrap) : bool = raise Eq
	in new_method eqeq
	end;
fun eq (p,wp,_) (q,wq,_) = get_eq () (wp p) (wq q);

(* new *)
fun new () = (pack () %, wrap'self, self'wrap);
end;



(* points -------------------------------------------------------------	*)

signature POINT =
sig
    type 'a ext
    type 'a subself = 'a ext Eqobj.subself
    type self = closed subself
    val wrap'self : self -> Root.wrap
    val self'wrap : Root.wrap -> self
    val unpack : 'a subself -> ('b -> 'b subself) * 'a
    val pack : int -> 'a -> 'a subself
    val pos : 'a subself Root.packet -> int
    val slide : 'a subself Root.packet -> int -> 'a subself Root.packet
    val overload_slide : ((Root.wrap -> int -> Root.wrap) ->
			  (Root.wrap -> int -> Root.wrap)) -> unit
    val dist : 'a subself Root.packet -> 'b subself Root.packet -> real
    val overload_dist : ((Root.wrap -> Root.wrap -> real) ->
			 (Root.wrap -> Root.wrap -> real)) -> unit
    val new : int -> self Root.packet
end;

structure Point :> POINT =
struct

(* types *)
type 'a ext = int * 'a;
type 'a subself = 'a ext Eqobj.subself;
type self = closed subself;

(* packing *)
fun unpack p = explode (Eqobj.unpack p);
fun repack f p = implode f (Eqobj.unpack p);
fun pack n ext = Eqobj.pack () (n, ext);

(* wrapping *)
val (Self,matches) = create (None : closed ext option);
fun wrap'self p =
	let val (f,ext) = Eqobj.unpack p in Eqobj.wrap'self (f (Self ext)) end;
fun self'wrap w =
	let val (f,ww) = Eqobj.unpack (Eqobj.self'wrap w)
	in case matches ww of Some ext => f ext | _ => raise Root.Wrap
	end;

(* messages *)
fun wpos p = let val (_, (x,_)) = Eqobj.unpack p in x end;
fun pos (p,_,_) = wpos p;
val (overload_slide,get_slide) =
	let fun wslide (p : Root.wrap) (n : int) : Root.wrap =
		wrap'self (repack (fn x => x+n) (self'wrap p))
	in new_method wslide
	end;
fun slide (p,w,u) n = (u (get_slide () (w p) n),w,u);
val (overload_dist,get_dist) =
	let fun pdist (p : Root.wrap) (q : Root.wrap) : real =
		real (wpos (self'wrap q) - wpos (self'wrap p));
	in new_method pdist
	end;
fun dist (p,wp,_) (q,wq,_) = get_dist () (wp p) (wq q);
val _ = (* specializing eq: *)
	let fun peq eqeq p q =
		wpos (self'wrap p) = wpos (self'wrap q)
		handle Root.Wrap => eqeq p q
	in Eqobj.overload_eq peq
	end;

(* new *)
fun new n = (pack n %, wrap'self, self'wrap);
end;


(* colored points  ----------------------------------------------------	*)

signature CPOINT =
sig
    type 'a ext
    type 'a subself = 'a ext Point.subself
    type self = closed subself
    val wrap'self : self -> Root.wrap
    val self'wrap : Root.wrap -> self
    val unpack : 'a subself -> ('b -> 'b subself) * 'a
    val pack : color * int -> 'a -> 'a subself
    val col : 'a subself Root.packet -> color
    val paint : 'a subself Root.packet -> color -> 'a subself Root.packet
    val overload_paint : ((Root.wrap -> color -> Root.wrap) ->
			  (Root.wrap -> color -> Root.wrap)) -> unit
    val new : int * color -> self Root.packet
end;

structure Cpoint :> CPOINT =
struct

(* types *)
type 'a ext = color * 'a;
type 'a subself = 'a ext Point.subself;
type self = closed subself;

(* packing *)
fun unpack p = explode (Point.unpack p);
fun repack f p = implode f (Point.unpack p);
fun pack (c,n) ext = Point.pack n (c, ext);

(* wrapping *)
val (Self,matches) = create (None : closed ext option);
fun wrap'self p =
	let val (f,ext) = Point.unpack p in Point.wrap'self (f (Self ext)) end;
fun self'wrap w =
	let val (f,ww) = Point.unpack (Point.self'wrap w)
	in case matches ww of Some ext => f ext | _ => raise Root.Wrap
	end;

(* messages *)
fun wcol p = let val (_,(c,_)) = Point.unpack p in c end;
fun col (p,w,u) = wcol p;
val (overload_paint,get_paint) =
	let fun wpaint (p : Root.wrap) (c:color) : Root.wrap =
		wrap'self (repack (fn _ => c) (self'wrap p));
	in new_method wpaint
	end;
fun paint (p,w,u) c = (u (get_paint () (w p) c),w,u);
val _ = (* overloads eq *)
	let fun cpeq peq p q =
		(wcol (self'wrap p) = wcol (self'wrap q) andalso peq p q)
		handle Root.Wrap => peq p q
	in Eqobj.overload_eq cpeq
      	end;

(* new *)
fun new (n,c) = (pack (c,n) %,wrap'self,self'wrap);
end;



(* 2d points ---------------------------------------------------------- *)

signature D2POINT =
sig
    type 'a ext
    type 'a subself = 'a ext Point.subself
    type self = closed subself
    val wrap'self : self -> Root.wrap
    val self'wrap : Root.wrap -> self
    val unpack : 'a subself -> ('b -> 'b subself) * 'a
    val pack : int * int -> 'a -> 'a subself
    val xpos : 'a subself Root.packet -> int
    val ypos : 'a subself Root.packet -> int
    val new : int * int -> self Root.packet
end;

structure D2point :> D2POINT =
struct

(* types *)
type 'a ext = int * 'a;
type 'a subself = 'a ext Point.subself;
type self = closed subself;

(* packing *)
fun unpack p = explode (Point.unpack p);
fun repack f p = implode f (Point.unpack p);
fun pack (y,x) ext = Point.pack x (y, ext);

(* wrapping *)
val (Self,matches) = create (None : closed ext option);
fun wrap'self p =
	let val (f,ext) = Point.unpack p in Point.wrap'self (f (Self ext)) end;
fun self'wrap w =
	let val (f,ww) = Point.unpack (Point.self'wrap w)
	in case matches ww of Some ext => f ext | _ => raise Root.Wrap
	end;

(* messages *)
fun wypos p = let val (_,(y,_)) = Point.unpack p in y end;
val xpos = Point.pos; 	(* alias *)
fun ypos (p,w,u) = wypos p;
val _ = (* overloads slide *)
	let fun d2slide pslide p n =
		pslide (wrap'self (repack (fn y => y+n) (self'wrap p))) n
		handle Root.Wrap => pslide p n
	in Point.overload_slide d2slide
      	end;
val _ = (* overloads dist *)
	let fun d2dist pdist p q =
		sqrt (sq (real (wypos (self'wrap q) - wypos (self'wrap p))) +
		      sq (pdist p q))
		handle Root.Wrap => pdist p q
	in Point.overload_dist d2dist
      	end;
val _ = (* overloads eq *)
	let fun d2eq peq p q =
		(wypos (self'wrap p) = wypos (self'wrap q) andalso peq p q)
		handle Root.Wrap => peq p q
	in Eqobj.overload_eq d2eq
      	end;

(* new *)
fun new (x,y) = (pack (y,x) %, wrap'self, self'wrap);
end; 




(* lines --------------------------------------------------------------- *)

signature MINIPOINT =
sig
    type 'a ext
    type 'a subself
    type self = closed subself
    val slide : 'a subself Root.packet -> int -> 'a subself Root.packet
    val dist : 'a subself Root.packet -> 'b subself Root.packet -> real
    val eq : 'a subself Root.packet -> 'b subself Root.packet -> bool
end;

signature LINE =
sig
    structure Param : MINIPOINT
    type 'a ext
    type 'a subself = 'a ext Eqobj.subself
    type self = closed subself
    val wrap'self : self -> Root.wrap
    val self'wrap : Root.wrap -> self
    val unpack : 'a subself -> ('b -> 'b subself) * 'a
    val pack : (Param.self Root.packet * Param.self Root.packet) ->
	       'a -> 'a subself
    val from : 'a subself Root.packet -> Param.self Root.packet
    val to   : 'a subself Root.packet -> Param.self Root.packet
    val shift : 'a subself Root.packet -> int -> 'a subself Root.packet
    val length : 'a subself Root.packet -> real
    val overload_length : ((Root.wrap -> real) ->
			   (Root.wrap -> real)) -> unit
    val overload_shift : ((Root.wrap -> int -> Root.wrap) ->
			  (Root.wrap -> int -> Root.wrap)) -> unit
    val new : Param.self Root.packet * Param.self Root.packet -> self Root.packet
end;

functor Line (structure P : MINIPOINT) :> LINE where Param = P =
struct

structure Param = P;

(* types *)
type 'a ext = (Param.self Root.packet * Param.self Root.packet) * 'a;
type 'a subself = 'a ext Eqobj.subself;
type self = closed subself;

(* packing *)
fun unpack p = explode (Eqobj.unpack p);
fun repack f p = implode f (Eqobj.unpack p);
fun pack ft ext = Eqobj.pack () (ft, ext);

(* wrapping *)
val (Self,matches) = create (None : closed ext option);
fun wrap'self p =
	let val (f,ext) = Eqobj.unpack p in Eqobj.wrap'self (f (Self ext)) end;
fun self'wrap w =
	let val (f,ww) = Eqobj.unpack (Eqobj.self'wrap w)
	in case matches ww of Some ext => f ext | _ => raise Root.Wrap
	end;

(* messages *)
fun wfrom l = let val (_,((f,_),_)) = Eqobj.unpack l in f end;
fun wto   l = let val (_,((_,t),_)) = Eqobj.unpack l in t end;
fun from (l,_,_) = wfrom l;
fun to   (l,_,_) = wto l;
val (overload_shift,get_shift) =
	let fun wshift (l : Root.wrap) (n:int) : Root.wrap =
		wrap'self (repack (fn (f,t) =>
				(Param.slide f n, Param.slide t n)) (self'wrap l));
	in new_method wshift
	end;
fun shift (l,w,u) n = (u (get_shift () (w l) n),w,u);
val (overload_length,get_length) =
	let fun wlength (l : Root.wrap) : real =
		Param.dist (wfrom (self'wrap l)) (wto (self'wrap l));
	in new_method wlength
	end;
fun length (l,w,u) = get_length () (w l);
val _ = (* overloads eq *)
	let fun leq eqeq p q =
		let val dp = self'wrap p
		    and dq = self'wrap q
		in Param.eq (wfrom dp) (wfrom dq) andalso Param.eq (wto dp) (wto dq)
		end
		handle Root.Wrap => eqeq p q
	in Eqobj.overload_eq leq
      	end;

(* new *)
fun new (f,t) = (pack (f,t) %, wrap'self, self'wrap)
end;

structure PointLine = Line (structure P =
				struct
				open Point;
				val eq = Eqobj.eq;
				end);

structure CpointLine = Line (structure P =
				struct
				open Cpoint;
 				val slide = Point.slide;
				val dist = Point.dist;
				val eq = Eqobj.eq;
				end);

structure D2pointLine = Line (structure P =
				struct
				open D2point;
 				val slide = Point.slide;
				val dist = Point.dist;
				val eq = Eqobj.eq;
				end);


