(* recursive-self, simple dispatch *)

use "../lib/smlprelude";


(* root objects -------------------------------------------------------	*)

open Closed;   (* closed datatype, % : closed *)

signature ROOT =
sig
    type 'a subself
    type self = closed subself
    val pack : unit -> 'a -> (unit -> 'a subself) -> 'a subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val new : unit -> self
end;

structure Root :> ROOT =
struct

(* types *)
type 'a subself = 'a;
type self = closed subself;

(* packing *)
fun unpack wob = (I,wob);
fun pack () ob self = ob;

(* new *)
fun new () = let fun self () = pack () % self in self () end;
end;
 

(* eqobjs ------------------------------------------------------------- *)

signature EQOBJ =
sig
    type 'a ext
    type 'a subself = 'a ext Root.subself
    type self = closed subself
    val pack : unit -> 'a -> (unit -> unit -> 'a subself) -> 'a subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val repack : (('a subself -> bool) -> ('a subself -> bool)) ->
		 'a subself -> 'a subself
    exception Eq;
    val eq : 'a subself -> 'a subself -> bool
    val new : unit -> self
end;

structure Eqobj :> EQOBJ =
struct

(* types *)
datatype 'a state = S of
		'a subself -> bool				(* eq    *)
withtype 'a ext = 'a state * 'a
and 'a subself = 'a ext Root.subself;
type self = closed subself;

(* messages *)
exception Eq;
fun eq p q = let val (_,(S e,_)) = Root.unpack p in e q end;

(* packing *)
fun unpack p = explode (Root.unpack p);
fun repack f p = implode (fn S m => S (f m)) (Root.unpack p);
fun pack () ext self =
	let fun eqeq w = raise Eq
	in Root.pack () (S eqeq, ext) (self ())
	end;

fun new () = let fun self () () = pack () % self in self () () end;
end;



(* points -------------------------------------------------------------	*)

signature POINT =
sig
    type 'a ext
    type 'a subself = 'a ext Eqobj.subself
    type self = closed subself
    val pack : int -> 'a -> (int -> unit -> unit -> 'a subself) -> 'a subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val repack : (int * (int -> 'a subself) * ('a subself -> real) ->
		  int * (int -> 'a subself) * ('a subself -> real)) ->
	         'a subself -> 'a subself
    val pos : 'a subself -> int
    val slide : 'a subself -> int -> 'a subself
    val dist : 'a subself -> 'a subself -> real
    val new : int -> self
end;

structure Point :> POINT =
struct

(* types *)
datatype 'a state = S of
		int *					(* pos	 *)
		(int -> 'a subself) * 			(* slide *)
		('a subself -> real)			(* dist  *)
withtype 'a ext = 'a state * 'a
and 'a subself = 'a ext Eqobj.subself;
type self = closed subself;

(* messages *)
fun pos p = let val (_,(S (n,_,_),_)) = Eqobj.unpack p in n end;
fun slide p n =	let val (_,(S (_,s,_),_)) = Eqobj.unpack p in s n end;
fun dist p q = let val (_,(S (_,_,d),_)) = Eqobj.unpack p in d q end;

(* packing *)
fun unpack p = explode (Eqobj.unpack p);
fun repack f p = implode (fn S m => S (f m)) (Eqobj.unpack p);
fun pack n ext self =
	let fun mslide k = self (n + k) () ()
	    fun mdist q = real (pos q - n)
	    fun peq eqeq q = n = pos q
	in Eqobj.repack peq (Eqobj.pack () (S (n,mslide,mdist), ext) (self n))
	end;

(* new *)
fun new n = let fun self n () () = pack n % self in self n () () end;
end;


(* colored points  ----------------------------------------------------	*)

signature CPOINT =
sig
    type 'a ext
    type 'a subself = 'a ext Point.subself
    type self = closed subself
    val pack : color * int ->
 	       'a -> (color -> int -> unit -> unit -> 'a subself) -> 'a subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val repack : (color * (color -> 'a subself) ->
		  color * (color -> 'a subself)) ->
	         'a subself -> 'a subself
    val col : 'a subself -> color
    val paint : 'a subself -> color -> 'a subself
    val new : int * color -> self
end;

structure Cpoint :> CPOINT = 
struct

(* types *)
datatype 'a state = S of
		color *					(* col	 *)
		(color -> 'a subself)			(* paint *)
withtype 'a ext = 'a state * 'a
and 'a subself = 'a ext Point.subself;
type self = closed subself;

(* messages *)
fun col p = let val (_,(S (c,_),_)) = Point.unpack p in c end;
fun paint p c =	let val (_,(S (_,t),_)) = Point.unpack p in t c end;

(* packing *)
fun unpack p = explode (Point.unpack p);
fun repack f p = implode (fn S m => S (f m)) (Point.unpack p);
fun pack (c,n) ext self =
	let fun mpaint c' = self c' n () ()
	    fun cpeq peq q = peq q andalso (c = col q)
	in Eqobj.repack cpeq (Point.pack n (S (c,mpaint), ext) (self c))
	end;

(* new *)
fun new (n,c) = let fun self c n () () = pack (c,n) % self	in self c n () () end;
end;



(* 2d points  ---------------------------------------------------------	*)

signature D2POINT =
sig
    type 'a ext
    type 'a subself = 'a ext Point.subself
    type self = closed subself
    val pack   : int * int ->
		 'a -> (int -> int -> unit -> unit -> 'a subself) -> 'a subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val repack : (int -> int) -> 'a subself -> 'a subself
    val xpos : 'a subself -> int
    val ypos : 'a subself -> int
    val new : int * int -> self
end;

structure D2point :> D2POINT =
struct

(* types *)
datatype state = S of
		int					(* ypos  *)
withtype 'a ext = state * 'a
and 'a subself = 'a ext Point.subself;
type self = closed subself;

(* messages *)
val xpos = Point.pos; (* alias *)
fun ypos p = let val (_,(S n,_)) = Point.unpack p in n end;

(* packing *)
fun unpack p = explode (Point.unpack p);
fun repack f p = implode (fn S m => S (f m)) (Point.unpack p);
fun pack (y,x) (ext:'a) self =
	let val mypos = y
	    fun mslide pslide k = self (y+k) (x+k) () ()
	    fun d2dist pdist q =
		sqrt (sq (real (ypos q - y)) + sq (pdist q))
	    fun d2peq peq q = peq q andalso (y = ypos q)
	in Eqobj.repack d2peq
			(Point.repack (fn (p,s,d) => (p,mslide s,d2dist d))
				      (Point.pack x (S mypos, ext) (self y)))
	end;

(* new *)
fun new (x,y) =	let fun self y x () () = pack (y,x) % self in self y x () () end;
end;



(* line --------------------------------------------------------------- *)

signature LINE =
sig
    type ('a,'b) ext
    type ('a,'b) subself = ('a,'b) ext Eqobj.subself
    type 'a self = ('a, closed) subself
    val pack   : 'a Point.subself * 'a Point.subself ->
		 'b -> 
		 ('a Point.subself * 'a Point.subself -> unit -> unit -> ('a,'b) subself) ->
		 ('a,'b) subself
    val unpack : ('a,'b) subself -> ('b -> ('a,'b) subself) * 'b
    val repack : ('a Point.subself * 
		  'a Point.subself * 
	          (int -> ('a,'b) subself) *
		  (unit -> real) ->
		  'a Point.subself *
		  'a Point.subself *
		  (int -> ('a,'b) subself) *
		  (unit -> real)) ->
		 ('a,'b) subself -> ('a,'b) subself
    val from : ('a,'b) subself -> 'a Point.subself
    val to : ('a,'b) subself -> 'a Point.subself
    val shift : ('a,'b) subself -> int -> ('a,'b) subself
    val length : ('a,'b) subself -> real
    val new : 'a Point.subself * 'a Point.subself -> 'a self
end;

structure Line :> LINE =
struct

(* types *)
datatype ('a,'b) state = S of
		'a Point.subself *			(* from *)
		'a Point.subself *			(* to   *)
		(int -> ('a,'b) subself)*		(* shift *)
		(unit -> real)                   	(* length *)
withtype ('a,'b) ext = ('a,'b) state * 'b
and ('a,'b) subself = ('a,'b) ext Eqobj.subself;
type 'a self = ('a,closed) subself;

(* messages *)
fun from l = let val (_,(S (f,_,_,_),_)) = Eqobj.unpack l in f end;
fun to l = let val (_,(S (_,t,_,_),_)) = Eqobj.unpack l in t end;
fun shift l n =	let val (_,(S (_,_,s,_),_)) = Eqobj.unpack l in s n end;
fun length l = let val (_,(S (_,_,_,le),_)) = Eqobj.unpack l in le () end;

(* packing *)
fun unpack p = explode (Eqobj.unpack p);
fun repack f p = implode (fn S m => S (f m)) (Eqobj.unpack p);
fun pack (f,t) ext self =
	let val mfr = f
	    val mto = t
	    fun msh n = self (Point.slide f n, Point.slide t n) () ()
	    fun mlen () = Point.dist f t
	    fun leq eqeq q = Eqobj.eq f (from q) andalso Eqobj.eq t (to q)
	in Eqobj.repack leq (Eqobj.pack () (S (mfr,mto,msh,mlen),ext) (self (f,t)))
	end;

(* new *)
fun new (f : 'a Point.subself, t : 'a Point.subself) : 'a self =
	let fun self ft () () = pack ft % self in self (f,t) () ()	end;
end;


